/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.registry.core;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

import org.ow2.petals.registry.api.Constants;
import org.ow2.petals.registry.api.Information;
import org.ow2.petals.registry.api.LocalRegistry;
import org.ow2.petals.registry.api.Query;
import org.ow2.petals.registry.api.Resource;
import org.ow2.petals.registry.api.context.Context;
import org.ow2.petals.registry.api.exception.LifeCycleException;
import org.ow2.petals.registry.api.exception.RegistryException;
import org.ow2.petals.registry.api.lifecycle.AbstractLifeCycle;
import org.ow2.petals.registry.api.lifecycle.LifeCycle;
import org.ow2.petals.registry.api.manager.IncomingManager;
import org.ow2.petals.registry.api.manager.OutgoingManager;
import org.ow2.petals.registry.api.manager.RemoteClientManager;
import org.ow2.petals.registry.api.repository.Repository;
import org.ow2.petals.registry.api.transport.MessageReceiver;
import org.ow2.petals.registry.api.util.ResourceUtil;

/**
 * The abstract class which represents a local registry.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @author Charles Casadei - eBM WebSourcing
 */
public abstract class AbstractLocalRegistry extends AbstractRegistry implements LocalRegistry {

    /**
     * The repository to be used to store resources
     */
    protected Repository repository = null;

    /**
     * The local manager to be used to manage local resources
     */
    protected IncomingManager incomingManager = null;

    /**
     * The remote manager is used to handle remote messages and to propagate
     * them or not...
     */
    protected OutgoingManager outgoingManager = null;

    /**
     * The message receiver is in charge of receiving messages from remote pairs
     */
    protected MessageReceiver messageReceiver = null;

    /**
     * The remote client manager
     */
    protected RemoteClientManager remoteClientManager = null;

    /**
     * The local revision of the registry. When a resource is added or deleted
     * the revision is incremented. This is used by all the registries to
     * synchronise their visions of the entries.
     */
    protected AtomicLong revision = null;

    protected ScheduledExecutorService executorService = null;

    public AbstractLocalRegistry() {
        super();
        this.lifeCycle = new LocalRegistryLifeCycle();
    }

    public void init() throws LifeCycleException {
        this.init(this.getContext());
    }

    @Override
    public String getName() {
        return this.getContext().getConfiguration().getName();
    }

    public LifeCycle getLifeCycle() {
        return this.lifeCycle;
    }

    public IncomingManager getIncomingManager() {
        return this.incomingManager;
    }

    public void setIncomingManager(IncomingManager incomingManager) throws LifeCycleException {
        if (!this.isNotInitialized()) {
            throw new LifeCycleException("Can not set the local manager in the state "
                    + this.getState());
        }
        this.incomingManager = incomingManager;
    }

    public OutgoingManager getOutgoingManager() {
        return this.outgoingManager;
    }

    public void setOutgoingManager(OutgoingManager outgoingManager) throws LifeCycleException {
        if (!this.isNotInitialized()) {
            throw new LifeCycleException("Can not set the remote manager in the state "
                    + this.getState());
        }
        this.outgoingManager = outgoingManager;
    }

    public Repository getRepository() {
        return this.repository;
    }

    public void setRepository(Repository repository) throws LifeCycleException {
        if (!this.isNotInitialized()) {
            throw new LifeCycleException("Can not set the repository in the state "
                    + this.getState());
        }
        this.repository = repository;
    }

    public MessageReceiver getMessageReceiver() {
        return this.messageReceiver;
    }

    public void setMessageReceiver(MessageReceiver messageReceiver) {
        this.messageReceiver = messageReceiver;
    }

    public RemoteClientManager getRemoteClientManager() {
        return this.remoteClientManager;
    }

    public void setRemoteClientManager(RemoteClientManager remoteClientManager)
            throws LifeCycleException {
        if (!this.isNotInitialized()) {
            throw new LifeCycleException("Can not set the client manager in the state "
                    + this.getState());
        }
        this.remoteClientManager = remoteClientManager;
    }

    public synchronized void pauseReceive() {
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Pause receive");
        }

        if (this.getIncomingManager() != null) {
            this.getIncomingManager().pause();
        }
    }

    public synchronized void resumeReceive() {
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Resume receive");
        }

        if (this.getIncomingManager() != null) {
            this.getIncomingManager().resume();
        }
    }

    public void clean() throws RegistryException {
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Ask to clean the repository");
        }
        this.getRepository().clean();
    }

    /**
     * {@inheritDoc}
     */
    public void cleanRemoteData() throws RegistryException {
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Ask to delete the data from other registries");
        }
        this.getRepository().keepOnly(this.getName() + Constants.SEPARATOR);
    }

    /**
     * Recover data from all the registries. TODO : SYnchronizing data means
     * that all is totally locked and that all operations are pending!!!
     * 
     * @throws RegistryException
     */
    public void synchronizeData() throws RegistryException {
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Synchronizing local data with other peers data...");
        }

        this.pauseReceive();

        try {
            try {
                // here is the list of already registered things (from all
                // nodes)
                List<Resource> oldResources = this.retrieveOldKnownResources();

                Map<String, Resource> oldies = new HashMap<String, Resource>();
                for (Resource resource : oldResources) {
                    oldies.put(resource.getPath(), resource);
                }

                // here is the list of local registered things (local node)
                List<Resource> localResources = this.retrieveLocalResources(oldResources);

                try {
                    // get all the resources from all the nodes
                    List<Resource> externalResources = this.retrieveExternalResources();

                    for (Resource localResource : localResources) {
                        // put and propagate to others
                        // FIXME : We do not need to put them again, just
                        // propagate data!
                        this.put(localResource.getPath(), localResource, true);
                    }

                    ListIterator<Resource> externalResourceIterator = externalResources
                            .listIterator();

                    while (externalResourceIterator.hasNext()) {
                        Resource externalResource = externalResourceIterator.next();
                        this.getRepository().put(externalResource.getPath(), externalResource,
                                false);

                        if (oldies.get(externalResource.getPath()) != null) {
                            if (this.logger.isDebugEnabled()) {
                                this.logger.debug("Found the external resource in oldies = "
                                        + externalResource.getPath());
                            }
                            oldies.remove(externalResource.getPath());
                        }

                        // if (oldResources.contains(externalResource)) {
                        // oldResources.remove(externalResource);
                        // }
                    }

                    /*
                     * if (oldResources.size() > 0) { for (Resource oldResource
                     * : oldResources) { if
                     * (!oldResource.getRootRegistry().equals(this.getName())) {
                     * this.delete(oldResource.getPath(), false); } } }
                     */

                    if (this.logger.isDebugEnabled()) {
                        this.logger.debug("Deleting some oldies which are obsoletes...");
                    }
                    Iterator<String> iter = oldies.keySet().iterator();
                    while (iter.hasNext()) {
                        Resource r = oldies.get(iter.next());
                        if ((r != null) && !r.getRootRegistry().equals(this.getName())) {
                            if (this.logger.isDebugEnabled()) {
                                this.logger.debug("Will delete " + r.getPath());
                            }
                            this.delete(r.getPath(), false);
                        }
                    }

                } catch (RegistryException e) {

                    // this is the rollback???
                    // for (Resource oldResource : oldResources) {
                    // this.getRepository().put(oldResource.getPath(),
                    // oldResource, false);
                    // }

                    throw e;
                }
            } catch (RegistryException e) {
                if (this.logger.isWarnEnabled()) {
                    this.logger.warn("Synchronization failure...");
                }
                throw e;
            }
        } finally {
            this.resumeReceive();
        }
    }

    private List<Resource> retrieveOldKnownResources() throws RegistryException {
        List<Resource> oldResources = this.getAll(Constants.ROOT_PATH, false);

        if (oldResources == null) {
            throw new RegistryException("Retrieved known resources == null");
        }

        return oldResources;
    }

    private List<Resource> retrieveLocalResources(List<Resource> oldKnownResources)
            throws RegistryException {
        List<Resource> localResources = new ArrayList<Resource>();

        if (oldKnownResources != null) {
            for (Resource oldResource : oldKnownResources) {
                if (oldResource.getRootRegistry().equals(this.getName())) {
                    localResources.add(oldResource);
                }
            }
        }

        return localResources;
    }

    private List<Resource> retrieveExternalResources() throws RegistryException {
        List<Resource> externalResources = this.outgoingManager.getAll(Constants.ROOT_PATH, this
                .createInformation());

        if (externalResources == null) {
            throw new RegistryException("Retrieved external resources == null");
        }

        return externalResources;
    }

    @Override
    public Resource doGet(String key, boolean b) throws RegistryException {
        // try to get locally
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Get the resource under " + key + " from local registry");
        }

        if (ResourceUtil.isDirectory(key)) {
            throw new RegistryException("Bad resource path '" + key + "'");
        }

        // 1. Try to get the resource locally
        Resource repositoryResource = this.getRepository().get(key, b);
        Resource remoteResource = null;

        // 2. Try to get the resource from remote servers
        if ((this.getOutgoingManager() != null) && (repositoryResource == null) && b) {
            this.logger
                    .debug("Resource has not been found locally, get it from remote registries...");
            remoteResource = this.getOutgoingManager().get(key, this.createInformation());
        }

        // 3. FIXME : The resource has been found somewhere, cache it locally
        // for
        // some time (depends on the resource timeout)
        if (b && (remoteResource != null)) {
            this.logger.debug("Resource has finally been found remotely, store it locally");
            try {
                this.getRepository().put(key, repositoryResource, b);
            } catch (RegistryException e) {
                this.logger.warn(e.getMessage());
                this.logger
                        .warn("Exception while caching to local registry after getting it from remote...");
            }
        }

        // resource can be null if it has not been found from local and remote
        // servers...
        return repositoryResource;
    }

    @Override
    public List<Resource> doGetAll(String key, boolean b) throws RegistryException {
        // try to get locally
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Get all the resources under " + key + " from local registry");
        }

        // check that the path is a folder first...
        if (!ResourceUtil.isDirectory(key)) {
            throw new RegistryException(key + " is not a directory, should have a trailing '"
                    + Constants.SEPARATOR + "'");
        }

        List<Resource> result = new ArrayList<Resource>();

        // 1. Try to get the resource locally
        List<Resource> localResources = this.getRepository().getAll(key, b);
        if (localResources != null) {
            result.addAll(localResources);
        }

        // 2. Try to get the resource from remote servers
        List<Resource> remoteResources = new ArrayList<Resource>();
        if (b && (this.getOutgoingManager() != null)) {
            if (this.logger.isDebugEnabled()) {
                this.logger.debug("Get all the resources under " + key + " from remote registries");
            }
            remoteResources = this.getOutgoingManager().getAll(key, this.createInformation());
        }

        return this.mergeResources(localResources, remoteResources);

        // TODO : merge the results ie the local oldies will be updated

        // 3. FIXME : The resource has been found somewhere, cache it locally
        // for
        // some time (depends on the resource timeout)
        // if (resource != null) {
        // try {
        // getRepository().put(key, resource);
        // } catch (Exception e) {
        // log
        // .warn("Exception while caching to local registry after getting it from remote...");
        // }
        // }

        // resource can be null if it has not been found from local and remote
        // servers...
    }

    @Override
    public boolean doPut(String key, Resource resource, boolean b) throws RegistryException {
        boolean result = false;

        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Put the resource to local repository under '" + key + "'");
        }

        if (ResourceUtil.isDirectory(key)) {
            throw new RegistryException("Bad path " + key);
        }

        resource.setPath(key);

        // save locally
        result = this.getRepository().put(key, resource, b);

        // propagate
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Propagate the resource to remote repositories");
        }

        if ((this.getOutgoingManager() != null) && result && b) {
            if (this.logger.isDebugEnabled()) {
                this.logger.debug("Propagating the PUT command on remote registries");
            }
            result = this.getOutgoingManager().put(key, resource, this.createInformation());
        }

        return result;
    }

    @Override
    public boolean doDelete(String key, boolean b) throws RegistryException {
        boolean result = false;

        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Delete the resource from local repository under " + key);
        }

        // 1. try to delete locally
        result = this.getRepository().delete(key, b);

        if ((this.getOutgoingManager() != null) && result && b) {
            // 2. delete remotely
            result = this.getOutgoingManager().delete(key, this.createInformation());
        } else {
            this.logger.debug("Local delete returned false, do not delete remotely...");
        }

        return result;
    }

    public List<Resource> query(Query query, boolean b) throws RegistryException {
        if (this.logger.isDebugEnabled()) {
            this.logger.debug("Query on local registry " + query);
        }

        // try to get locally
        List<Resource> repo = new ArrayList<Resource>();
        List<Resource> repoTmp = this.getRepository().query(query, b);
        if (repoTmp != null) {
            if (this.logger.isDebugEnabled()) {
                this.logger.debug(repoTmp.size() + " entry(ies) returned from repository");
            }
            repo.addAll(repoTmp);
        } else {
            if (this.logger.isDebugEnabled()) {
                this.logger.debug("No entries found from repository");
            }
        }

        List<Resource> remote = new LinkedList<Resource>();
        if ((this.getOutgoingManager() != null) && b) {
            if (this.logger.isDebugEnabled()) {
                this.logger.debug("Let's query remote repositories");
            }
            List<Resource> tmp = this.getOutgoingManager().query(query, this.createInformation());
            if (tmp != null) {
                if (this.logger.isDebugEnabled()) {
                    this.logger.debug(tmp.size() + " entry(ies) returned from remote repositories");
                }
                remote.addAll(tmp);
            } else {
                if (this.logger.isDebugEnabled()) {
                    this.logger.debug("No entries found from remote repositories");
                }
            }
        }

        return this.mergeResources(repo, remote);
    }

    protected List<Resource> mergeResources(List<Resource> repository, List<Resource> remote) {

        List<Resource> result = new LinkedList<Resource>();

        Map<String, Resource> map = new HashMap<String, Resource>();
        if (repository != null) {
            for (Resource resource : repository) {
                if (this.logger.isDebugEnabled()) {
                    this.logger.debug("Add local repository resource to local list " + resource);
                }
                map.put(resource.getPath(), resource);
            }
        }

        if (remote != null) {
            for (Resource resource : remote) {
                // TODO : check the time, location and more...
                if (map.get(resource.getPath()) == null) {
                    if (this.logger.isDebugEnabled()) {
                        this.logger.debug("Add remote resource to local list " + resource);
                    }
                    map.put(resource.getPath(), resource);
                } else {
                    Resource cached = map.get(resource.getPath());
                    if ((this.getName().equals(resource.getRootRegistry()) && (resource
                            .getCreationTime() >= cached.getCreationTime()))) {
                        map.put(resource.getPath(), resource);
                    }
                }
            }
        }

        result.addAll(map.values());
        return result;
    }

    private Information createInformation() {
        Information information = new Information();
        information.setSender(this.getName());
        information.setTime(System.currentTimeMillis());
        return information;
    }

    protected class LocalRegistryLifeCycle extends AbstractLifeCycle {
        // TODO handle exceptions and rollback things if problems occur

        public LocalRegistryLifeCycle() {
            super();
        }

        @Override
        protected void doInit(Context context) throws LifeCycleException {
            if (AbstractLocalRegistry.this.logger.isDebugEnabled()) {
                AbstractLocalRegistry.this.logger.info("AbstractLifeCycle doInit()");
            }

            context.getWorkspace().init();

            AbstractLocalRegistry.this.revision = new AtomicLong(0L);

            AbstractLocalRegistry.this.getRepository().init(context);

            if (AbstractLocalRegistry.this.getIncomingManager() != null) {
                AbstractLocalRegistry.this.getIncomingManager().init(context);
            }

            if (AbstractLocalRegistry.this.getOutgoingManager() != null) {
                AbstractLocalRegistry.this.getOutgoingManager().init(context);
            }

            // optional components
            if (AbstractLocalRegistry.this.getRemoteClientManager() != null) {
                AbstractLocalRegistry.this.getRemoteClientManager().init(context);
            }
        }

        @Override
        protected void doStart() throws LifeCycleException {
            if (AbstractLocalRegistry.this.logger.isDebugEnabled()) {
                AbstractLocalRegistry.this.logger.debug("AbstractLifeCycle doStart()");
            }

            AbstractLocalRegistry.this.getRepository().start();

            if (AbstractLocalRegistry.this.getIncomingManager() != null) {
                AbstractLocalRegistry.this.getIncomingManager().start();
            }

            if (AbstractLocalRegistry.this.getOutgoingManager() != null) {
                AbstractLocalRegistry.this.getOutgoingManager().start();
            }

            // start receiving messages...
            if (AbstractLocalRegistry.this.getMessageReceiver() != null) {
                AbstractLocalRegistry.this.getMessageReceiver().startReceiving();
            }

            // once all is started, start optional components

            // optional components
            if (AbstractLocalRegistry.this.getRemoteClientManager() != null) {
                AbstractLocalRegistry.this.getRemoteClientManager().start();
            }

            String updatePeriodProperty = null;
            if ((AbstractLocalRegistry.this.context != null)
                    && (AbstractLocalRegistry.this.context.getConfiguration() != null)) {
                updatePeriodProperty = AbstractLocalRegistry.this.context.getConfiguration()
                        .getRegistrySynchroPeriod();
            }

            boolean hasUpdatePeriodProperty = updatePeriodProperty != null;
            boolean hasValidUpdatePeriod = hasUpdatePeriodProperty ? Long
                    .parseLong(updatePeriodProperty) > 0 : false;
            String registryMode = AbstractLocalRegistry.this.context.getConfiguration()
                    .getRegistryMode();

            boolean isStandalone = org.ow2.petals.registry.api.config.Constants.MODE.STANDALONE
                    .toString().equalsIgnoreCase(registryMode);
            boolean isMaster = org.ow2.petals.registry.api.config.Constants.MODE.MASTER.toString()
                    .equalsIgnoreCase(registryMode);
            boolean isSlave = org.ow2.petals.registry.api.config.Constants.MODE.SLAVE.toString()
                    .equalsIgnoreCase(registryMode);
            boolean isPeer = org.ow2.petals.registry.api.config.Constants.MODE.PEER.toString()
                    .equalsIgnoreCase(registryMode);

            if (AbstractLocalRegistry.this.logger.isDebugEnabled()) {
                AbstractLocalRegistry.this.logger.debug("Registry mode is set to '" + registryMode
                        + "'");
            }

            if (!isStandalone && !isMaster && (isPeer || isSlave) && hasValidUpdatePeriod) {
                Thread
                        .setDefaultUncaughtExceptionHandler(new LocalRegistryUncaughtExceptionHandler());

                long updatePeriod = Long.parseLong(updatePeriodProperty);
                AbstractLocalRegistry.this.executorService = Executors.newScheduledThreadPool(1);
                // Schedule the update at a fixed rate, 30s after petals has
                // been launched
                AbstractLocalRegistry.this.executorService.scheduleAtFixedRate(
                        new LocalRegistryUpdater(), 30, updatePeriod, TimeUnit.SECONDS);
            }
        }

        @Override
        protected void doStop() throws LifeCycleException {
            if (AbstractLocalRegistry.this.logger.isDebugEnabled()) {
                AbstractLocalRegistry.this.logger.debug("AbstractLifeCycle doStop()");
            }

            // optional components
            if (AbstractLocalRegistry.this.getRemoteClientManager() != null) {
                try {
                    AbstractLocalRegistry.this.getRemoteClientManager().stop();
                } catch (Exception e) {
                    this.log.warn(e.getMessage());
                }
            }

            // stop receiving message first
            if (AbstractLocalRegistry.this.getMessageReceiver() != null) {
                AbstractLocalRegistry.this.getMessageReceiver().stopReceiving();
            }

            // TODO : handle exceptions when a stop fails, all must be
            // stopped!!!!
            if (AbstractLocalRegistry.this.getOutgoingManager() != null) {
                try {
                    AbstractLocalRegistry.this.getOutgoingManager().stop();
                } catch (Exception e) {
                    this.log.warn(e.getMessage());
                }
            }

            if (AbstractLocalRegistry.this.getIncomingManager() != null) {
                try {
                    AbstractLocalRegistry.this.getIncomingManager().stop();
                } catch (Exception e) {
                    this.log.warn(e.getMessage());
                }
            }

            try {
                AbstractLocalRegistry.this.getRepository().clean();
            } catch (Exception e) {
                this.log.warn(e.getMessage());
            }

            if (AbstractLocalRegistry.this.executorService != null) {
                AbstractLocalRegistry.this.executorService.shutdown();
            }
        }
    }

    protected class LocalRegistryUpdater implements Runnable {

        public void run() {
            if (AbstractLocalRegistry.this.logger.isDebugEnabled()) {
                AbstractLocalRegistry.this.logger
                        .debug("Update the local registry with remote informations");
            }
            try {
                AbstractLocalRegistry.this.synchronizeData();
            } catch (Throwable ex) {
                AbstractLocalRegistry.this.logger
                        .error("Local Registry : Can't synchronize with remote registries : " + ex);
            }
        }
    }

    protected class LocalRegistryUncaughtExceptionHandler implements
            Thread.UncaughtExceptionHandler {

        public void uncaughtException(Thread th, Throwable ex) {
            AbstractLocalRegistry.this.logger.error("Error in thread : " + th.getName() + " : "
                    + ex);
        }
    }
}
