/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.registry.core.factory;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;

import org.ow2.petals.registry.api.LocalRegistry;
import org.ow2.petals.registry.api.RemoteRegistry;
import org.ow2.petals.registry.api.config.Configuration;
import org.ow2.petals.registry.api.config.RemoteConfiguration;
import org.ow2.petals.registry.api.context.Context;
import org.ow2.petals.registry.api.context.ContextFactory;
import org.ow2.petals.registry.api.exception.RegistryException;
import org.ow2.petals.registry.api.manager.IncomingManager;
import org.ow2.petals.registry.api.manager.OutgoingManager;
import org.ow2.petals.registry.api.manager.RemoteClientManager;
import org.ow2.petals.registry.api.repository.Repository;
import org.ow2.petals.registry.api.transport.MessageReceiver;
import org.ow2.petals.registry.api.transport.MessageSender;
import org.ow2.petals.registry.api.util.Assert;
import org.ow2.petals.registry.api.util.Loader;
import org.ow2.petals.registry.core.LocalRegistryImpl;
import org.ow2.petals.registry.core.RemoteRegistryImpl;

/**
 * @author Christophe HAMERLING - eBM WebSourcing
 * 
 */
public class RegistryFactory {

    private static RegistryFactory instance;

    public synchronized static RegistryFactory getInstance() {
        if (instance == null) {
            instance = new RegistryFactory();
        }
        return instance;
    }

    private RegistryFactory() {
    }

    /**
     * Load a new local registry instance from the given configuration file
     * 
     * @return
     */
    public LocalRegistry load(File configFile) throws RegistryException {
        Context ctx = ContextFactory.getInstance().createContext(configFile);
        return this.loadLocal(ctx);
    }

    /**
     * Load the registry from a configuration bean.
     * 
     * @param configuration
     * @return
     * @throws RegistryException
     */
    public LocalRegistry loadLocal(Configuration configuration) throws RegistryException {
        Context ctx = ContextFactory.getInstance().createContext(configuration);
        return this.loadLocal(ctx);
    }

    /**
     * 
     * @param registryName
     *            the name of the resgitry. This name MUST be unique in the
     *            registry network. If not, the registry will be rejected
     * @param config
     * @return
     */
    public LocalRegistry loadLocal(Context context) throws RegistryException {
        Configuration config = context.getConfiguration();
        Assert.notNull(config, "The configuration can not be null");
        Assert.notNull(config.getName(), "The registry name can not be null");

        String registryName = config.getName();

        LocalRegistry registry = new LocalRegistryImpl();
        registry.setName(registryName);
        registry.setContext(context);

        IncomingManager lm = null;
        if ((config.getIncomingManagerClassName() != null)
                && (config.getIncomingManagerClassName().trim().length() > 0)) {
            lm = Loader.loadLocalManager(config.getIncomingManagerClassName());
            Assert.notNull(lm);
            lm.setLocalRegistry(registry);
            registry.setIncomingManager(lm);
        }

        if ((config.getOutgoingManagerClassName() != null)
                && (config.getOutgoingManagerClassName().trim().length() > 0)) {
            OutgoingManager rm = Loader.loadRemoteManager(config.getOutgoingManagerClassName());
            Assert.notNull(rm);
            rm.setLocalRegistry(registry);
            registry.setOutgoingManager(rm);
        }

        // Repository
        Repository r = RepositoryFactory.load(config);

        registry.setRepository(r);
        registry.setName(registryName);

        // optional components
        if ((config.getMessageReceiverClassName() != null)
                && (config.getMessageReceiverClassName().trim().length() > 0)) {
            MessageReceiver receiver = Loader.loadMessageReceiver(config.getMessageReceiverClassName());
            receiver.setLocalManager(lm);
            registry.setMessageReceiver(receiver);
        }

        if (config.getClientManagerClassName() != null) {
            RemoteClientManager clientManager = Loader.loadRemoteClientManager(config.getClientManagerClassName());
            clientManager.setLocalRegistry(registry);
            registry.setRemoteClientManager(clientManager);
        }

        return registry;
    }

    /**
     * Create a remote registry instance. This just create all the stuff needed
     * to communicate from the local registry to the remote one
     * 
     * @return
     */
    public RemoteRegistry loadRemote(RemoteConfiguration configuration) throws RegistryException {
        Assert.notNull(configuration);
        URI uri = null;
        try {
            uri = new URI(configuration.getUri());
        } catch (URISyntaxException e) {
            throw new RegistryException(e);
        }

        RemoteRegistry remote = new RemoteRegistryImpl();
        MessageSender messageSender = Loader.loadMessageSender(configuration.getMessageSenderClassName());
        messageSender.setURI(uri);
        remote.setMessageSender(messageSender);
        remote.getMessageSender().setURI(uri);
        remote.setName(configuration.getName());
        remote.setURI(uri);

        return remote;
    }
}
